/*
 * Copyright (C) 1995, 1996, 1997, and 1998 WIDE Project.
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *    This product includes software developed by WIDE Project and
 *    its contributors.
 * 4. Neither the name of the project nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE PROJECT AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE PROJECT OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#define ATM_PCR
#define NATMP2P 2

#include <sys/param.h>
#include <sys/malloc.h>
#include <sys/socket.h>
#include <sys/sockio.h>
#include <sys/ioctl.h>
#include <sys/errno.h>
#include <sys/sysctl.h>

#include <net/if.h>
#include <net/if_dl.h>
#include <net/route.h>

#include <net/if_atm.h>

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <strings.h>
#include <ctype.h>
#include <err.h>


/* interface list retrieval functions related declarations */
#define	IFL_ANY		1	/* Any interfaces */
#define	IFL_UPONLY	2	/* 'UP' interfaces only */
#define	IFL_DOWNONLY	3	/* 'DOWN' interfaces only */

struct iflist {
    char* buf;			/* if_msg buffer addr */
    char* next;			/* current working addr */
    char* end;			/* next addr of end of if_msg */
    char name[IFNAMSIZ+1];	/* current interface name */
};

struct iflist *open_iflist();
char* get_iflist(struct iflist*, int);
void close_iflist(struct iflist *);


void usage();
void setvc(int, char**);
void parse_dot_notation(int, char**, struct aptpifreq*);
void parse_slash_notation(int, char**, struct aptpifreq*);
void showvc(int, char**);
void showstatus(int, char*, int);


main(argc, argv)
int argc;
char *argv[];
{
    if (argc < 2) {
	usage();
    } else if (argc == 2) {
	showvc(argc, argv);
    } else {
	setvc(argc, argv);
    }
    exit(0);
}

void
usage()
{
    fprintf(stderr,
	    "usage: apconfig p2p-if atm-if:VPI/VCI "
					"[aal5 [llc]]"
#ifdef ATM_PCR
	    " [pcr NUM|ubr]"
#endif
	    "\n"
	    "       apconfig p2p-if atm-if:a.b.c.d.e.f.g.h\n"
	    "       apconfig [-a|-au|-ad]\n");
    exit(1);
}


#define	DECVAL(c)	( (c) - '0' )
#define	HEXVAL(c)	( isdigit(c) ? ((c) - '0') :			\
			  (isupper(c) ? ((c) - 'A') : ((c) - 'a')) + 10 )

void
setvc(argc, argv)
int argc;
char *argv[];
{
    struct aptpifreq ptpr;
/*    struct aptppar req;*/
    int s;

    bzero(&ptpr, sizeof(ptpr));
/*    bzero(&req, sizeof(req));
    ptpr.req = &req;*/
    strncpy(ptpr.aptpifr_name, argv[1], IFNAMSIZ);

    if (index(argv[2], '.')) {
	/* apconfig ap0 en0:a.b.c.d.e.f.g.h */
	parse_dot_notation(argc, argv, &ptpr);
    } else if (index(argv[2], '/')) {
	/* apconfig ap0 en0:0/100 aal5 llc pcr 100 */
	parse_slash_notation(argc, argv, &ptpr);
    } else {
	usage();
    }
    /* don't we need to check parameters here? */

    if ((s = socket(AF_INET, SOCK_DGRAM, 0)) < 0)    
	err(1, "cannot open socket");
    if( ioctl(s, SIOCAPTPS_VC, (caddr_t)&ptpr) < 0)
	err(1, "failed to set parameters");
    close(s);
}


/* to parse "apconfig ap0 en0:a.b.c.d.e.f.g.h" */
void
parse_dot_notation(argc, argv, ptpr)
int		argc;
char**		argv;
struct aptpifreq*	ptpr;
{
    char *p;
    int i;
    int tmp;

    if ( (p = index(argv[2], ':')) == NULL || (p - argv[2]) > IFNAMSIZ )
	usage();
    *p++ = '\0';
    strcpy(ptpr->atmdev, argv[2]);

    for (i = 0; i < 8; i++) {
	if (!isxdigit(*p))
	    usage();
	tmp = HEXVAL(*p);
	p++;
	if (isxdigit(*p)) {
	    tmp = tmp * 16 + HEXVAL(*p);
	    p++;
	}
	ptpr->aph.atm_ph[i] = tmp;
	if ((i < 7  && *p != '.') || (i == 7 && *p != '\0'))
	    usage();
	if (*p == '.')
	    p++;
    }
}


/* to parse "apconfig ap0 en0:0/100 aal5 llc pcr 100" */
void
parse_slash_notation(argc, argv, ptpr)
int		argc;
char**		argv;
struct aptpifreq*	ptpr;
{
    char *p;
    int i;
    int flag, vpi, vci;
#ifdef ATM_PCR
    int pcr, ubr;
#endif

    if ( (p = index(argv[2], ':')) == NULL || (p - argv[2]) > IFNAMSIZ )
	usage();
    *p++ = '\0';
    strcpy(ptpr->atmdev, argv[2]);

    if (!isdigit(*p))
	usage();
    for (vpi = 0; isdigit(*p); p++) {
	vpi = vpi * 10 + DECVAL(*p);
    }
    if (*p++ != '/')
	usage();
    for (vci = 0; isdigit(*p); p++) {
	vci = vci * 10 + DECVAL(*p);
    }
    if (*p != '\0')
	usage();

    flag = pcr = ubr = 0;
    for (i = 3; i < argc; i++) {
	if (strcasecmp("aal5", argv[i]) == 0) {
	    flag |= ATM_PH_AAL5;
	} else if (strcasecmp("llc", argv[i]) == 0) {
	    if (!(flag & ATM_PH_AAL5))
		fprintf(stderr, "\"llc\" must appear after \"aal5\".\n");
	    flag |= ATM_PH_LLCSNAP;
	}
#ifdef ATM_PCR
	else if (strcasecmp("ubr", argv[i]) == 0) {
	    if (ubr < 0)	/* if pcr option is also specified, */
		usage();	/* show usage and exit. */
	    ubr = 1;		/* it indicates ubr option is specified. */

	    pcr = 0;
	} else if (strcasecmp("pcr", argv[i]) == 0) {
	    i++;
	    if (i >= argc)
		usage();
	    if (ubr > 0)	/* if ubr option is also specified, */
		usage();	/* show usage and exit. */
	    ubr = -1;		/* it indicates pcr option is specified. */

	    pcr = atoi(argv[i]);
	}
#endif
	else {
	    usage();
	}
    }

    bzero(&ptpr->aph, sizeof(ptpr->aph));
    ATM_PH_FLAGS(&ptpr->aph) = flag;
    ATM_PH_VPI(&ptpr->aph) = vpi;
    ATM_PH_SETVCI(&ptpr->aph, vci);
#ifdef ATM_PCR
    ATM_SET_PCR(&(ptpr->aph), pcr);
#endif
}


void
showvc(argc, argv)
int argc;
char *argv[];
{
    int s;
    int updown;
    char*  ifname  = NULL;
    struct iflist* stat;

    if ((s = socket(AF_INET, SOCK_DGRAM, 0)) < 0)
	err(1, "cannot open socket");

    if (argv[1][0] == '-') {
	if (strcmp(argv[1], "-a") == 0) {
	    updown = IFL_ANY;
	} else if (strcmp(argv[1], "-au") == 0) {
	    updown = IFL_UPONLY;
	} else if (strcmp(argv[1], "-ad") == 0) {
	    updown = IFL_DOWNONLY;
	} else {
	    usage();
	}
	stat = open_iflist();
	if (stat == NULL)
	    errx(1, "out of memory");
	while ((ifname = get_iflist(stat, updown)) != NULL) {
	    showstatus(s, ifname, 0);
	}	
	close_iflist(stat);
    } else {
	showstatus(s, argv[1], 1);
    }

    close(s);
}



void
showstatus(s, ifname, force)
int s;
char* ifname;
int force;
{
    struct aptpifreq ptpr;
/*    struct aptppar req;*/
    char   atmifname[IFNAMSIZ+1];
    static int first = 1;

    bzero(&ptpr, sizeof(ptpr));
/*    bzero(&req, sizeof(req));
    ptpr.req = &req;*/
    strncpy(ptpr.aptpifr_name, ifname, IFNAMSIZ);
    if (ioctl(s, SIOCAPTPG_VC, (caddr_t)&ptpr) < 0) {
	if (!force)
	    return;
	switch (errno) {
	 case ENXIO:
	    errx(1, "interface %s does not exist.", ifname);
	 case EINVAL:
	 default:
	    errx(1, "%s is not an atm p2p interface.", ifname);
	}
	/*NOTREACHED*/
    }

    strncpy(atmifname, ptpr.atmdev, IFNAMSIZ);
    atmifname[IFNAMSIZ] = '\0';

    if (first) {
	printf(" IF  ATM  VPI/VCI/ PCR\n");
	first = 0;
    }

    if (*(ptpr.atmdev) == NULL) {
	printf("%4s   -    -/  -/   -\n", ifname);
	return;
    }

    printf("%4s%4s  %3d/%3d/",
	   ifname, atmifname, ATM_PH_VPI(&ptpr.aph), ATM_PH_VCI(&ptpr.aph));

    if (ATM_PH_PCR(&ptpr.aph) != 0) {
	printf("%4d", ATM_PH_PCR(&ptpr.aph));
    } else {
	printf(" UBR");
    }

    if (ATM_PH_FLAGS(&ptpr.aph)) {
	printf(" ");
	if (ATM_PH_FLAGS(&ptpr.aph) & ATM_PH_AAL5)
	    printf(" AAL5");
	if (ATM_PH_FLAGS(&ptpr.aph) & ATM_PH_LLCSNAP)
	    printf(" LLC/SNAP");
    }

    printf("\n");
}


int mib[6] = { CTL_NET, PF_ROUTE, 0, 0, NET_RT_IFLIST, 0 };

struct iflist *
open_iflist()
{
    size_t needed;
    struct iflist* ifl;

    ifl = malloc(sizeof(struct iflist));
    if (ifl == NULL)
	errx(1, "out of memory");

    if (sysctl(mib, 6, NULL, &needed, NULL, 0) < 0)
	errx(1, "failed to get interface list");
    if ((ifl->buf = malloc(needed)) == NULL)
	errx(1, "out of memory");
    if (sysctl(mib, 6, ifl->buf, &needed, NULL, 0) < 0)
	errx(1, "failed to get interface list");

    ifl->next = ifl->buf;
    ifl->end  = ifl->buf + needed;

    return ifl;
}

char*
get_iflist(ifl, updown)
struct iflist*	ifl;
int		updown;
{
    struct if_msghdr *ifm;
    struct sockaddr_dl *sdl;

    while (ifl->next < ifl->end) {
	ifm = (struct if_msghdr*)ifl->next;
	ifl->next += ifm->ifm_msglen;

	if (ifm->ifm_type == RTM_NEWADDR)
	    continue;
	if (ifm->ifm_type != RTM_IFINFO)
	    errx(1, "interface list parsing error");
	if ((updown == IFL_UPONLY   && !(ifm->ifm_flags & IFF_UP)) ||
	    (updown == IFL_DOWNONLY &&  (ifm->ifm_flags & IFF_UP))) {
	    continue;
	}
	sdl = (struct sockaddr_dl *)(ifm + 1);
	strncpy(ifl->name, sdl->sdl_data, sdl->sdl_nlen);
	ifl->name[sdl->sdl_nlen] = '\0';
	return ifl->name;
    }

    return NULL;
}

void
close_iflist(ifl)
struct iflist* ifl;
{
    free(ifl);
}
