/*
 * Copyright (C) 1995, 1996, 1997, and 1998 WIDE Project.
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *    This product includes software developed by WIDE Project and
 *    its contributors.
 * 4. Neither the name of the project nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE PROJECT AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE PROJECT OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#include <sys/param.h>
#include <sys/socket.h>
#ifdef ADVAPI
#include <sys/uio.h>
#endif
#include <sys/ioctl.h>

#include <net/if.h>
#include <netinet/in.h>
#include <netinet/ip6.h>
#include <netinet/icmp6.h>
#include <unistd.h>
#include <stdio.h>
#include <err.h>
#include <errno.h>
#include <string.h>
#include <stdlib.h>

#define ALLROUTER "ff02::2"

void usage __P((void));
int sendit __P((int, char *, int, struct sockaddr_in6 *, int));
void interface_up __P((char *));

#define DAD_RETRY_COUNT	3	/* should be net.inet6.ip6.dad_count? */
static int dad_retry = 0;

int
main(argc, argv)
	int argc;
	char *argv[];
{
	struct sockaddr_in6 to;
	struct icmp6_hdr *icp;
	u_int hlim = 255;
	u_char outpack[sizeof(struct icmp6_hdr)];
	u_short index;
	int s, i, cc = sizeof(struct icmp6_hdr);
	
	if (argc != 2)
		usage();

	index = (u_short)if_nametoindex(argv[1]);
	if (index == 0)
		usage();

	interface_up(argv[1]);

	memset(&to, 0, sizeof(struct sockaddr_in6));
	if (inet_pton(AF_INET6, ALLROUTER, &to.sin6_addr.s6_addr) < 0)
		exit(1);
	to.sin6_len = sizeof(struct sockaddr_in6);
	to.sin6_family = AF_INET6;
	
	if ((s = socket(AF_INET6, SOCK_RAW, IPPROTO_ICMPV6)) < 0)
		err(1, "socket");

	if (setsockopt( s, IPPROTO_IPV6, IPV6_MULTICAST_HOPS, &hlim,
			sizeof(hlim)) == -1)
		err(1, "IPV6_MULTICAST_HOPS");
	
	/* fill ICMP header */
	icp = (struct icmp6_hdr *)outpack;
	icp->icmp6_type = ND_ROUTER_SOLICIT;
	icp->icmp6_code = 0;
	icp->icmp6_cksum = 0;

retry:
	i = sendit(s, (char *)outpack, cc, &to, index);

	if (i < 0) {
		if (errno == EADDRNOTAVAIL && 0 < dad_retry--) {
			sleep(1);
			goto retry;
		}
		perror("rtsol: solicitation");
	} else if (i != cc) {
		fprintf(stderr, "rtsol: short write (wrote %d of be %d)\n",
			i, cc);
	}

	close(s);
	exit(0);
}

void
usage()
{
	(void)fprintf(stderr, "usage: rtsol ifname\n");
	exit(1);
}

int
sendit(s, p, len, sock, ifindex)
	int s;
	char *p;
	int len;
	struct sockaddr_in6 *sock;
	int ifindex;
{
#ifndef ADVAPI
	sock->sin6_addr.u6_addr.u6_addr16[1] = htons(ifindex);
	return sendto(s, p, len, 0, (struct sockaddr *)sock, sock->sin6_len);
#else
	struct msghdr m;
	struct cmsghdr *cm;
	struct iovec iov[2];
	u_char cmsgbuf[256];
	struct in6_pktinfo *pi;

	m.msg_name = (caddr_t)sock;
	m.msg_namelen = sock->sin6_len;
	iov[0].iov_base = p;
	iov[0].iov_len = len;
	m.msg_iov = iov;
	m.msg_iovlen = 1;
	if (!ifindex) {
		m.msg_control = NULL;
		m.msg_controllen = 0;
	} else {
		memset(cmsgbuf, 0, sizeof(cmsgbuf));
		cm = (struct cmsghdr *)cmsgbuf;
		m.msg_control = (caddr_t)cm;
		m.msg_controllen = CMSG_SPACE(sizeof(struct in6_pktinfo));

		cm->cmsg_len = CMSG_LEN(sizeof(struct in6_pktinfo));
		cm->cmsg_level = IPPROTO_IPV6;
		cm->cmsg_type = IPV6_PKTINFO;
		pi = (struct in6_pktinfo *)CMSG_DATA(cm);
		memset(&pi->ipi6_addr, 0, sizeof(pi->ipi6_addr));	/*XXX*/
		pi->ipi6_ifindex = ifindex;
	}
	return sendmsg(s, &m, 0);
#endif
}

void
interface_up(name)
	char *name;
{
	int s;
	struct ifreq ifr;

	if ((s = socket(AF_INET6, SOCK_DGRAM, 0)) < 0) {
		perror("socket");
		exit(1);
	}
	strncpy(ifr.ifr_name, name, sizeof(ifr.ifr_name));

	if (ioctl(s, SIOCGIFFLAGS, (caddr_t)&ifr) < 0)
		perror("ioctl(SIOCGIFFLAGS)");
	if (ifr.ifr_flags & IFF_UP)
		return;

	ifr.ifr_flags |= IFF_UP;
	if (ioctl(s, SIOCSIFFLAGS, (caddr_t)&ifr) < 0)
		perror("ioctl(SIOCSIFFLAGS)");
	dad_retry = DAD_RETRY_COUNT;
}
