/*
 * Copyright (C) 1995, 1996, 1997, and 1998 WIDE Project.
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *    This product includes software developed by WIDE Project and
 *    its contributors.
 * 4. Neither the name of the project nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE PROJECT AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE PROJECT OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

%{
#include <sys/types.h>
#include <sys/param.h>
#include <sys/socket.h>
#include <netkey/keyv2.h>
#include <netkey/keydb.h>
#include <netkey/key_debug.h>
#include <netinet/in.h>
#include <netinet6/in6.h>
#include <netinet6/ipsec.h>

#include <stdlib.h>
#include <limits.h>
#include <string.h>
#include <ctype.h>
#include <unistd.h>
#include <errno.h>
#include "vchar.h"
#include "y.tab.h"

#define ATOX(c) \
  (isdigit(c) ? (c - '0') : (isupper(c) ? (c - 'A' + 10) : (c - 'a' + 10) ))

#define KECHO \
{ \
	if (f_debug) {printf("<%d>", yy_start); ECHO ; printf("\n"); } \
	if (yytext[0] == '\t') yytext[0] = ' '; \
	strcat(cmdarg, yytext); \
}

int lineno = 1;
u_char cmdarg[8192]; /* XXX: BUFSIZ is the better ? */
struct _ptrlist {
	void *ptr;
	struct _ptrlist *next;
} ptrlist;

extern u_char m_buf[BUFSIZ];
extern u_int m_len;
extern int f_debug;

void yyerror(char *s);
void addptrlist(void *);
void delptrlist(void);

extern void parse_init(void);

%}

/* common section */
nl		\n
ws		[ \t]+
digit		[0-9]
letter		[0-9A-Za-z]
hexdigit	[0-9A-Fa-f]
octet		(([01]?{digit}?{digit})|((2([0-4]{digit}))|(25[0-5])))
special		[()\.\-+:\|\?\*,]
dot		\.
hyphen		\-
colon		\:
slash		\/
bcl		\{
ecl		\}
blcl		\[
elcl		\]
percent		\%
semi		\;
usec		{dot}{digit}{1,6}

comment		\#.*
ccomment	"/*"
bstring		\<[^>]*\>
qstring		\"[^"]*\"
number		{digit}+
hexpair		{hexdigit}{hexdigit}
hex8string	0[xX]{hexdigit}{1,8}
hexstrings	{hexpair}+({hyphen}{hexpair}+)+
octetstring	{octet}({dot}{octet})+
	/* ipaddress	({digit}{1,3}{dot}){3}{digit}{1,3} */
ipaddress	{ipv4addr}|{ipv6addr}
ipv4addr	{digit}{1,3}({dot}{digit}{1,3}){0,3}
ipv6addr	{hexdigit}{0,4}({colon}{hexdigit}{0,4}){2,7}
ipaddrmask	{slash}{digit}{1,3}
ipaddrport	{blcl}{number}{elcl}
keyword		{letter}{letter}+
name		{letter}(({letter}|{digit}|{hyphen})*({letter}|{digit}))*
hostname	{name}(({dot}{name})+{dot}?)?
ipseclevel	none|avail|use|require|unique
ipsecpolicy	{ipseclevel}({colon}{ipseclevel}){3}

%%

add		{ KECHO; yylval.num = SADB_ADD; return(ADD); }
delete		{ KECHO; yylval.num = SADB_DELETE; return(DELETE); }
get		{ KECHO; yylval.num = SADB_GET; return(GET); }
flush		{ KECHO; yylval.num = SADB_FLUSH; return(FLUSH); }
dump		{ KECHO; yylval.num = SADB_DUMP; return(DUMP); }
dumpscript	{ KECHO; yylval.num = SADB_DUMP; return(DUMPSCRIPT); }

	/* for management SPD */
spdadd		{ KECHO; yylval.num = SADB_X_SPDADD; return(SPDADD); }
spddelete	{ KECHO; yylval.num = SADB_X_SPDDELETE; return(SPDDELETE); }
spddump		{ KECHO; yylval.num = SADB_X_SPDDUMP; return(SPDDUMP); }
spdflush	{ KECHO; yylval.num = SADB_X_SPDFLUSH; return(SPDFLUSH); }
{hyphen}P	{ KECHO; return(F_POLICY); } /* for SPD */
discard		{ KECHO; return(PL_DISCARD); }
bypass		{ KECHO; return(PL_BYPASS); }
{ipsecpolicy}	{
			u_char *p = yytext;
			static u_char n[SADB_PROTO_MAX];
			int i;

			KECHO;

			for (i = 0; i < SADB_PROTO_MAX; i++) {
				while (*p != ':' && *p != NULL) p++;
				*p++ = NULL;

				if (strcmp(yytext, "none") == 0)
					n[i] = IPSEC_LEVEL_NONE;
				else
				if (strcmp(yytext, "avail") == 0)
					n[i] = IPSEC_LEVEL_AVAIL;
				else
				if (strcmp(yytext, "use") == 0)
					n[i] = IPSEC_LEVEL_USE;
				else
				if (strcmp(yytext, "require") == 0)
					n[i] = IPSEC_LEVEL_REQUIRE;
				else
				if (strcmp(yytext, "unique") == 0)
					n[i] = IPSEC_LEVEL_UNIQUE;

				yytext = p;
			}

			yylval.val.buf = n;
			yylval.val.len = sizeof(n);
			return(PL_IPSEC);
		}

	/* flags */
{hyphen}p	{ KECHO; return(F_PROTOCOL); }
{hyphen}r	{ KECHO; return(F_REPLAY); }
{hyphen}E	{ KECHO; return(F_ENC); }
{hyphen}A	{ KECHO; return(F_AUTH); }

	/* protocols */
tcp		{ KECHO; yylval.num = IPPROTO_TCP; return(PR_TCP); }
udp		{ KECHO; yylval.num = IPPROTO_UDP; return(PR_UDP); }
ah		{ KECHO; yylval.num = SADB_SATYPE_AH; return(PR_AH); }
esp		{ KECHO; yylval.num = SADB_SATYPE_ESP; return(PR_ESP); }
ah-old		{ KECHO; yylval.num = SADB_SATYPE_AH; return(PR_AHOLD); }
esp-old		{ KECHO; yylval.num = SADB_SATYPE_ESP; return(PR_ESPOLD); }

	/* authentication alogorithm */
hmac-md5	{ KECHO; yylval.num = SADB_AALG_MD5HMAC; return(ALG_AUTH); }
hmac-sha1	{ KECHO; yylval.num = SADB_AALG_SHA1HMAC; return(ALG_AUTH); }
keyed-md5	{ KECHO; yylval.num = SADB_AALG_MD5; return(ALG_AUTH); }
null		{ KECHO; yylval.num = SADB_AALG_NULL; return(ALG_AUTH); }

	/* encryption alogorithm */
des-cbc		{ KECHO; yylval.num = SADB_EALG_DESCBC; return(ALG_ENC); }
3des-cbc	{ KECHO; yylval.num = SADB_EALG_3DESCBC; return(ALG_ENC); }
simple		{ KECHO; yylval.num = SADB_EALG_NULL; return(ALG_ENC); }
blowfish-cbc	{ KECHO; yylval.num = SADB_EALG_BLOWFISHCBC; return(ALG_ENC); }
cast128-cbc	{ KECHO; yylval.num = SADB_EALG_CAST128CBC; return(ALG_ENC); }
rc5-cbc		{ KECHO; yylval.num = SADB_EALG_RC5CBC; return(ALG_ENC); }
des-deriv	{ KECHO; yylval.num = SADB_EALG_DESCBC; return(ALG_ENCOLD); }
3des-deriv	{ KECHO; yylval.num = SADB_EALG_3DESCBC; return(ALG_ENCOLD); }

	/* extension */
random		{ KECHO; yylval.num = SADB_X_EXT_PRAND; return(EXTENSION); }
seq		{ KECHO; yylval.num = SADB_X_EXT_PSEQ; return(EXTENSION); }
zero		{ KECHO; yylval.num = SADB_X_EXT_PZERO; return(EXTENSION); }

	/* ... */
{ws}		{ KECHO; }
{nl}		{ lineno++; }
{comment}
{semi}		{ return(EOT); }

	/* parameter */
{ipv4addr}	{
			struct sockaddr_in *in;
			u_int sa_len;

			KECHO;

			sa_len = sizeof(struct sockaddr_in);
			if ((in = (struct sockaddr_in *)malloc(sa_len)) == 0)
				return(-1);
			addptrlist(in);

			memset((caddr_t)in, 0, sa_len);
			in->sin_family = PF_INET;
			in->sin_len = sa_len;
			(void)inet_pton(PF_INET, yytext, &in->sin_addr);

			yylval.val.len = sa_len;
			yylval.val.buf = (caddr_t)in;

			return(IP4_ADDRESS);
		}

{ipv6addr}	{
			struct sockaddr_in6 *in6;
			u_int sa_len;

			KECHO;

			sa_len = sizeof(struct sockaddr_in6);
			if ((in6 = (struct sockaddr_in6 *)malloc(sa_len)) == 0)
				return(-1);
			addptrlist(in6);

			memset((caddr_t)in6, 0, sa_len);
			in6->sin6_family = PF_INET6;
			in6->sin6_len = sa_len;
			(void)inet_pton(PF_INET6, yytext, &in6->sin6_addr);

			yylval.val.len = sa_len;
			yylval.val.buf = (caddr_t)in6;

			return(IP6_ADDRESS);
		}

{ipaddrmask}	{
			KECHO;
			yytext++;
			yylval.num = atoi(yytext);
			return(PREFIX);
		}

{ipaddrport}	{
			u_char *p = yytext;
			KECHO;
			while (*++p != ']') ;
			*p = NULL;
			yytext++;
			yylval.num = atoi(yytext);
			return(PORT);
		}

{number}	{
			char *bp;

			KECHO;
			yylval.num = strtol(yytext, &bp, 10);
			return(NUMBER);
		}

{hexstrings}	{
			char *bp, *yp;

			KECHO; 

			yylval.val.len = (yyleng - 2 ) / 2;
			if ((yylval.val.buf = malloc(yylval.val.len)) == 0)
				return(-1);
			addptrlist(yylval.val.buf);

			bp = yylval.val.buf;

			for (yp = yytext + 2; *yp; yp += 2, bp++)
				*bp = (ATOX(yp[0]) << 4) | ATOX(yp[1]);

			return(HEXSTRINGS);
		}

{hex8string}	{
			char *bp;

			KECHO;
			yylval.num = strtol(yytext, &bp, 16);
			return(NUMBER);
		}

{qstring}	{
			u_char *p = yytext;
			KECHO;
			while (*++p != '"') ;
			*p = NULL;
			yytext++;
			yylval.val.len = yyleng-2;
			yylval.val.buf = yytext;
			return(STRINGS);
		}

.		{ yyerror("Syntax error"); }

%%

void
yyerror(char *s)
{
	printf("line %d: %s at [%s]\n", lineno, s, yytext);
}

int
parse(fp)
	FILE **fp;
{
	yyin = *fp;

	parse_init();

	memset((caddr_t)&ptrlist, 0, sizeof(ptrlist));

	if(yyparse()) {
		printf("parse failed.\n");
		return(-1);
	}

	return(0);
}

void
addptrlist(ptr)
	void *ptr;
{
	struct _ptrlist *pl = &ptrlist;

	while (pl->next) pl = pl->next;

	pl->next = malloc(sizeof(struct _ptrlist));
	pl = pl->next;
	pl->ptr = ptr;
	pl->next = 0;

	return;
}

void
delptrlist()
{
	struct _ptrlist *pl = &ptrlist;

	if (ptrlist.ptr == 0) return;

	do {
		(void)free(pl->ptr);
		pl = pl->next;
	} while (pl);

	return;
}
